/* Copyright 2003, Jochen Hoenicke
 * 
 * Der Algorithmus ist wie folgt:

Zuerst werden alle Rotationen der Teile die in der Quader passen
berechnet und in der Datenstruktur "compiled" abgelegt.  Die Teile
sind dabei in einer verketteten Liste.  Von dem x-frmigen Teil werden
alle symmetrischen Stellungen entfernt, so dass am Ende nur 12
Rotationen brig bleiben, von denen allerdings 4 noch eine Symmetrie
haben.  Fr diese symmetrischen Stellungen wird dann die Symmetrie
ber ein asymmetrisches Teil entfernt (das zunchst in asym kompiliert
wird).  In dieser Datenstruktur werden zu jedem Bit die Rotationen
gespeichert, die dieses bit als niederwertigstes haben.

Als erstes wird dann das x-frmige Teil in placeFirst gelegt (der
berhmt-berchtigte X-Trick), je nachdem ob es symmetrisch ist oder
nicht wird dann asym oder das nicht reduzierte Teil in die verkettete
Liste gehngt.  Der X-Trick folgt der These den Such-Baum schmal zu
halten, denn es gibt nur 12 Mglichkeiten, whrend es fr Bit 0 ber
80 Mglichkeiten gibt es zu fllen.

Weil die Reihenfolge der bits bestimmt in welcher Reihenfolge der
Quader gefllt wird, werden anschlieend die bits so geshufflet dass
es in jedem Schritt mglichst wenige Rotationen gibt.  Dazu berechnet
reorder die mittlere Anzahl der Rotationen fr jedes bit ber
Wahrscheinlichkeiten.  Anschliessend werden alle Teile entsprechend
der neuen Bit-Reihenfolge transponiert.

Nun luft endlich die rekursive solve-Methode (zu der es ein etwas
schnelleres Assembler-Pendant gibt).  Diese benutzt eine Hashtabelle
um hnliche Stellungen (d.h. gleiche Teile und gleiche Felder belegt)
zu erkennen und dort abzukrzen.  Ab einer gewissen Rekursionsstufe
luft dann nur noch die schnelle Assembler-Version, die keine
Hashtabelle benutzt.

Als weiterer Trick werden ab Bit Nr. 32 nur noch das hi-word der
Rotationen gespeicher, denn das lo-word ist sowieso leer.  Die
Assembler-Version hat entsprechen zwei Teile, einen falls das lo-word
schon gefllt ist, und eine bei der das nicht ist.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ctpuz.h"

#define NUM_ROTATIONS  24
#define FIELD_SIZE     60
#define MAX_PIECE_SIZE 10

#define X_PIECE 10
#define ASYMMETRIC_PIECE 6
#define NUM_PIECES 12

#define coord2bit(x,y,z)  ((x)*12+(y)*3+(z)+4)
#define shrink_subs(x)  ((x) >> 10)

typedef int coord[3];
typedef coord piece[MAX_PIECE_SIZE];

unsigned char bitordering[64];

/* The compiled pieces are held in a linked list.
 * For each firstbit value rotations hold the rotated pieces.
 */
typedef struct compiled_piece {
  struct compiled_piece *next;
  long long             *rots[32];
  long                  *rots_hi[32];
  int                    nr;
} compiled_piece;

typedef struct hashed_pos {
  long long pos;
  int solutions;
  int substates;
} hashed_pos;

static __inline__ unsigned long
log2 (unsigned long word)
{
  __asm__ ("bsfl %1,%0"
:	   "=r" (word)
:	   "r" (word));
  return word;
}

hashed_pos *hashtable;
#define hashcode(pos,used) \
    (((pos>>4) ^ (pos >> 13) ^ (pos >> 23) ^ (pos >> 40) ^ used) & (HASH_SIZE-1))

piece ctpieces[NUM_PIECES] = {
  {
    {0,0,0}, {1,0,0}, {0,1,0}, {0,2,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {1,0,0}, {2,0,0}, {1,1,0}, {0,2,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {0,1,0}, {0,2,0}, {0,3,0}, {1,3,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {0,1,0}, {1,1,0}, {0,2,0}, {0,3,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {0,1,0}, {1,1,0}, {1,1,1}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {2,0,0}, {1,1,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {0,1,0}, {1,1,0}, {2,1,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {0,1,0}, {1,1,0}, {1,2,0}, {2,2,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {0,1,0}, {0,1,1}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {1,1,0}, {2,1,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {1,0,0}, {0,1,0}, {1,1,0}, {2,1,0}, {1,2,0}, {-1,-1,-1}
  }, {
    {0,0,0}, {1,0,0}, {2,0,0}, {3,0,0}, {0,1,0}, {2,1,0}, {-1,-1,-1}
  }
};

compiled_piece compiled[NUM_PIECES];
compiled_piece asym;

compiled_piece reordered[NUM_PIECES];
compiled_piece *first_piece;

#ifdef STATISTICS
static void dumpBoard(long long board, long long piece) {
  char bchar[4] = {'.', 'p', 'X', '*'};
  int z, y, x;
  for (z = 2; z >= 0; z--) {
    for (y = 3; y >= 0; y--) {
      for (x = 0; x < 5; x++) {
	int i = bitordering[coord2bit(x,y,z)];
	int b = (board >> i) & 1;
	int p = (piece >> i) & 1;
	printf("%c ", bchar[b*2+p]);
      }
      printf("   ");
      for (x = 0; x < 5; x++) {
	int i = bitordering[coord2bit(x,y,z)];
	printf(" %2d", i);
      }
      printf("\n");
    }
    printf("\n");
  }
}
#endif

void rotate(coord c, int rot, coord dest) {
  switch (rot / 4) {
  case 0:
    dest[0] = c[0];
    dest[1] = c[1];
    dest[2] = c[2];
    break;
  case 1:
    dest[0] = c[0];
    dest[1] = c[2];
    dest[2] = -c[1];
    break;
  case 2:
    dest[0] = c[1];
    dest[1] = c[0];
    dest[2] = -c[2];
    break;
  case 3:
    dest[0] = c[1];
    dest[1] = c[2];
    dest[2] = c[0];
    break;
  case 4:
    dest[0] = c[2];
    dest[1] = c[1];
    dest[2] = -c[0];
    break;
  case 5:
    dest[0] = c[2];
    dest[1] = c[0];
    dest[2] = c[1];
    break;
  }
  if (rot & 1) {
    dest[0] = -dest[0];
    dest[2] = -dest[2];
  }
  if (rot & 2) {
    dest[1] = -dest[1];
    dest[2] = -dest[2];
  }
}

long long rotate_and_shift(piece p, int rot, int shift)
{
  int i;
  long long bitvec = 0LL;

  for (i = 0; i < MAX_PIECE_SIZE; i++) {
    coord dest;
    if (p[i][0] == -1)
      break;
    rotate(p[i], rot, dest);
    dest[0] += shift % 5;
    dest[1] += (shift / 5) % 4;
    dest[2] += (shift / 20);
    if (dest[0] < 0 || dest[0] >= 5
	|| dest[1] < 0 || dest[1] >= 4
	|| dest[2] < 0 || dest[2] >= 3) {
      /* piece doesn't fit.. */
      return 0LL;
    }
    bitvec |= 1LL << coord2bit(dest[0], dest[1], dest[2]);
  }
  return bitvec;
}

int better(long long rot1, long long rot2) {
  /* Better means 
   *   1. the lowest bit nr is higher.
   *   2. the value is bigger.
   */
  return (rot1 & -rot1) > (rot2 & -rot2)
    || ((rot1 & -rot1) == (rot2 & -rot2)
	&& rot1 >= rot2);
}

long long mirror(long long rotation) {
  long long rot = 0LL;
  int x,y,z;

  for (x=0; x < 5; x++) {
    for (y = 0; y < 4; y++) {
      for (z = 0; z < 3; z++) {
	if (rotation & (1LL << coord2bit(x,y,z))) {
	  rot |= 1LL << coord2bit(4-x,  y,2-z);
	}
      }
    }
  }
  return rot;
}

int best_full_rotation(long long rotation, int axes) {
  int x,y,z;
  long long rot0, rot1, rot2, rot3;

  if (axes == 2) {
    /* calculate the four rotations of the piece, but with
     * reversed bit order.
     */
    rot0 = rot1 = rot2 = rot3 = 0LL;
    for (x=0; x < 5; x++) {
      for (y = 0; y < 4; y++) {
	for (z = 0; z < 3; z++) {
	  if (rotation & (1LL << coord2bit(x,y,z))) {
	    rot0 |= 1LL << (63-coord2bit(  x,  y,  z));
	    rot1 |= 1LL << (63-coord2bit(4-x,3-y,  z));
	    rot2 |= 1LL << (63-coord2bit(  x,3-y,2-z));
	    rot3 |= 1LL << (63-coord2bit(4-x,  y,2-z));
	  }
	}
      }
    }
    return rot0 >= rot1
      &&   rot0 >= rot2
      &&   rot0 >= rot3;
    
  } else {
    /* calculate the four rotations of the piece, but with
     * reversed bit order.
     */
    rot0 = rot1 = rot2 = rot3 = 0LL;
    for (x=0; x < 5; x++) {
      for (y = 0; y < 4; y++) {
	for (z = 0; z < 3; z++) {
	  if (rotation & (1LL << coord2bit(x,y,z))) {
	    rot0 |= 1LL << (63 - coord2bit(  x,  y,  z));
	    rot3 |= 1LL << (63 - coord2bit(4-x,  y,2-z));
	  }
	}
      }
    }
    /* We want the lowest bit of current rotation be higher than for any
     * other of the four rotation.  Since we reversed the bits, we want
     * the highest bit of rot0 be smaller than for any other rot.
     * We can use normal < operation.
     */
    return rot0 <= rot3;
  }
}


void compile(int pnr, int reduce_symm, compiled_piece *dest) {
  int rot, shift, i, j;
  long long rotpiece;
  int numrot[64];
  int totalrot = 0;
  int hirot = 0;
  long long rots[64][24*6];
  long long *comprots;
  long *comprots_hi;

  dest->nr = pnr;
  
  for (shift = 0; shift < 64; shift++)
    numrot[shift] = 0;

  for (rot = 0; rot < NUM_ROTATIONS; rot++) {
    for (shift = 0; shift < FIELD_SIZE; shift++) {
      rotpiece = rotate_and_shift(ctpieces[pnr], rot, shift);
      if (!rotpiece)
	continue;

      if (reduce_symm) {
	/* Only take this piece if it is better than the 
	 * other 3 rotations with respect to the whole cuboid.
	 *
	 * Better means 
	 *   1. the lowest bit nr is higher.
	 *   2. the value is bigger.
	 */
	if (!best_full_rotation(rotpiece, reduce_symm))
	  continue;
      }

      /* put piece at right position in array */
      for (i = 0; i < FIELD_SIZE; i++) {
	if ((rotpiece & 1LL << i))
	  break;
      }

      for (j = 0; j < numrot[i]; j++) {
	if (rots[i][j] == rotpiece)
	  /* Rotation was already seen */
	  goto next_rotation;
      }

      rots[i][numrot[i]] = rotpiece;
      numrot[i]++;
      totalrot++;
      if (i > 31)
	hirot++;
      if (numrot[i] > 24 * 6) {
	printf("Too many rotations for %d,%d!",pnr,i);
	exit(0);
      }

      /*printf("%2d,%2d,%2d: %16Lx\n", pnr, i, rot, rotpiece);*/
      /*  dumpBoard(0, rotpiece);*/
    }
  next_rotation:
  }

  printf("Piece %d, %d rotations\n", pnr, totalrot);

  comprots = malloc(sizeof(long long) * (totalrot - hirot + 28)
		    + sizeof(long) * (hirot + 32));
  for (i = 4; i < 32; i++) {
    dest->rots[i] = comprots;
    for (j = 0; j < numrot[i]; j++)
      *comprots++ = rots[i][j];
    *comprots++ = 1LL;
  }

  comprots_hi = (long *) comprots;
  for (i = 32; i < 64; i++) {
    dest->rots_hi[i-32] = comprots_hi;
    for (j = 0; j < numrot[i]; j++)
      *comprots_hi++ = rots[i][j] >> 32;
    *comprots_hi++ = 0LL;
  }
}

int solutions;


int used;
long long rotpieces[NUM_PIECES];
#ifdef COUNT_PLACED
unsigned int placedP[NUM_PIECES];
#endif
#ifdef COUNT_SAVED
int hits = 0;
unsigned int hitsP[NUM_PIECES] = {0};
unsigned int savedP[NUM_PIECES] = {0};
long long int saved = 0;
#endif
int placed = 0;

#ifdef COUNT_STEPS
int ctr[20] = {0,};
#endif

#if 0
static void dumpSolution(int pieces) {
  char bchar[] = "0123456789AB....";
  int z, y, x, i;
  for (z = 2; z >= 0; z--) {
    for (y = 3; y >= 0; y--) {
      for (x = 0; x < 5; x++) {
	long long mask = 1LL << bitordering[coord2bit(x,y,z)];
	for (i = 0; i < pieces; i++) {
	  if ((rotpieces[i] & mask))
	    break;
	}
	printf("%c ", i == pieces? '.' : bchar[i]);
      }
      printf("\n");
    }
    printf("\n");
  }
}
#endif

#ifdef STATISTICS
void statistics() {
#if defined(COUNT_PLACED) || defined(COUNT_SAVED) || defined(COUNT_STEPS)
    int i;
#endif
    printf("placed: %u ", placed);
#ifdef COUNT_PLACED
    for (i = 0; i < NUM_PIECES; i++) {
	printf("%u+", placedP[i]);
    }
#endif
    printf("\n");
#ifdef COUNT_SAVED
    printf("Hits: %d  splitted:", hits);
    for (i = 0; i < NUM_PIECES; i++) {
	printf("%u+", hitsP[i]);
    }
    printf("\n");
    printf("Saved: %lld  splitted:", saved);
    for (i = 0; i < NUM_PIECES; i++) {
	printf("%u, ", savedP[i]);
    }
    printf("\n");
#endif
    
#ifdef COUNT_STEPS
    printf("Counters: ");
    for (i = 0; i < 20; i++) {
	printf("%u, ", ctr[i]);
    }
    printf("\n");
#endif
}
#endif

extern void solve_asm(long long board, int firstbit, int depth);

/* This is the C version of the recursive solve method.  There is also
 * an assembler version that is used for the last rotations, where speed
 * is important.  The C version differs in that it keeps a hashtable and
 * that it can't handle rot_hi.
 */
void solve(long long board, int firstbit, int depth) {
  int i;
  int solved = 1;
  int hc, oldplaced, oldsolutions;
  struct compiled_piece **ppiece;

  for (i = firstbit; i < FIELD_SIZE; i++) {
    if (!(board & (1LL << i))) {
      firstbit = i;
      break;
    }
  }
  if (firstbit > 31) {
    solve_asm(board, firstbit, depth);
    return;
  }

  ppiece = &first_piece;
  while (*ppiece) {
    struct compiled_piece *piece;
    long long *rots;

    piece = *ppiece;
    *ppiece = piece->next;
    used ^= 1 << piece->nr;
    rots = piece->rots[firstbit];

    solved = 0;
    while (!(*(long*)rots & 1)) {
      long long piecemask = *rots++;

      if ((board & piecemask) == 0LL) {

	long long nboard;
#ifdef COUNT_PLACED
	placedP[depth]++;
#endif
	placed++;
	rotpieces[depth] = piecemask;

	oldsolutions = solutions;
	oldplaced = placed;

	nboard = board | piecemask;
	hc = hashcode(nboard, used);
	if (hashtable[hc].pos == nboard) {
	  int subs;
	  solutions += hashtable[hc].solutions;

	  subs = hashtable[hc].substates;
	  placed += subs;
#ifdef COUNT_SAVED
	  saved += subs;
	  savedP[depth] += subs;
	  hitsP[depth]++;
	  hits++;
#endif
	} else {	  
	  if (depth < HASH_DEPTH) {
	    solve(nboard, firstbit+1, depth + 1);
	  } else {
#if NUM_PIECES == HASH_DEPTH + 1
	    dumpSolution(depth + 1);
	    solutions++;
#else
	    solve_asm(nboard, firstbit+1, depth + 1);
#endif
	  }

	  if (hashtable[hc].substates < placed - oldplaced)
	  {
	    hashtable[hc].pos = nboard;
	    hashtable[hc].solutions = solutions - oldsolutions;
	    hashtable[hc].substates = placed - oldplaced;
	  }
	}
#if 0
	if (solutions == oldsolutions && (placed - oldplaced) > 100000) {
	  dumpSolution(depth + 1);
	  printf("has no solutions but %d pieces placed\n", placed - oldplaced);
	}
#endif
#ifdef STATISTICS
#if STATISTICS >= 2
	if (depth < STATISTICS /* && solutions > oldsolutions*/) {
	  printf("%d: %2d %4d solutions ; %7d solutions total\n", 
		 depth, piece->nr, solutions - oldsolutions, solutions);
	  dumpBoard(board, piecemask);
	  statistics();
	}
#endif
#endif
      }
    }
    used ^= 1 << piece->nr;
    *ppiece = piece;
    ppiece = &piece->next;
  }
}

long long transpose(long long piece) {
    long long result = 0LL;
    int i;
    for (i = 0; i < 64; i++) {
	if (piece & (1ULL << i)) {
	    result |= 1ULL << bitordering[i];
	}
    }
    return result;
}

void transpose_piece(struct compiled_piece *piece, 
		     struct compiled_piece *dest) {
    int totalrot = 0;
    int hirot = 0;
    long long rots[64][24*6];
    long long *prots;
    long *prots_hi;
    int numrot[64];
    int i, j;
    for (i = 4; i < 64; i++)
	numrot[i] = 0;

    for (i = 4; i < 32; i++) {
	prots = piece->rots[i];
	while (*prots != 1ULL) {
	    long long rot = transpose(*prots);
	    for (j = 4; j < 64; j++)
		if (rot & (1ULL << j))
		    break;
	    rots[j][numrot[j]++] = rot;
	    totalrot++;
	    if (j >= 32)
		hirot++;
	    prots++;
	}
    }
    for (i = 32; i < 64; i++) {
	prots_hi = piece->rots_hi[i-32];
	while (*prots_hi != 0LL) {
	    long long rot = transpose((long long) *prots_hi << 32);
	    for (j = 4; j < 64; j++)
		if (rot & (1ULL << j))
		    break;
	    rots[j][numrot[j]++] = rot;
	    totalrot++;
	    if (j >= 32)
		hirot++;
	    prots_hi++;
	}
    }
    
    dest->nr = piece->nr;
    prots = malloc(sizeof(long long) * (totalrot - hirot + 28)
		   + sizeof(long) * (hirot + 32));
    for (i = 4; i < 32; i++) {
	dest->rots[i] = prots;
	for (j = 0; j < numrot[i]; j++)
	    *prots++ = rots[i][j];
	*prots++ = 1LL;
    }
    
    prots_hi = (long *) prots;
    for (i = 32; i < 64; i++) {
	dest->rots_hi[i-32] = prots_hi;
	for (j = 0; j < numrot[i]; j++)
	    *prots_hi++ = rots[i][j] >> 32;
	*prots_hi++ = 0L;
    }
}

long long reorder(long long board, struct compiled_piece *piece_list) {
    double cube_freeness[64];
    double rotations[64];
    double rotations_used[64];
    unsigned char usedbit[64];
    int i, j;
    struct compiled_piece *p;
    struct compiled_piece **last;
    int bit = 0;

    for (i = 0; i < 64; i++) {
	if (!(board & (1ULL << i))) {
	    cube_freeness[i] = 1.0;
	    usedbit[i] = 0;
	} else {
	    bitordering[i] = bit++;
	    cube_freeness[i] = 0.0;
	    usedbit[i] = 1;
	}
    }

/*#define DEBUG_BITORDERING	*/
#ifdef DEBUG_BITORDERING	
    printf ("Bitordering:");
#endif
    for (; bit < 64; bit++) { 
	int best;
	double bestrot;
	for (j = 4; j < 64; j++) {
	    rotations[j] = 0.0;
	    rotations_used[j] = 0.0;
	}
	for (p = piece_list; p; p = p->next) {
	    for (i = 4; i < 32; i++) {
		long long *rots = p->rots[i];
		while (*rots != 1LL) {
		    double weight = 1.0;
		    for (j = i; j < 64; j++) {
			if (*rots & (1ULL << j)) {
			    weight *= cube_freeness[j];
			}
		    }
		    /* weight is the probability, that this rotation
		     * is still usable
		     */
		    for (j = i; j < 64; j++) {
			if (*rots & (1ULL << j)) {
			    rotations[j] += weight;
			}
		    }
		    rots++;
		}
	    }
	    for (i = 32; i < 64; i++) {
		long *rots_hi = p->rots_hi[i-32];
		while (*rots_hi != 0) {
		    double weight = 1.0;
		    for (j = i; j < 64; j++) {
			if (*rots_hi & (1UL << (j-32))) {
			    weight *= cube_freeness[j];
			}
		    }
		    for (j = i; j < 64; j++) {
			if (*rots_hi & (1UL << (j-32))) {
			    rotations[j] += weight;
			}
		    }
		    rots_hi++;
		}
	    }
	}
	best = 0;
	bestrot = 1000.0;
	for (i = 4; i < 64; i++) {
	    if (!usedbit[i]
		&& (rotations[i] * cube_freeness[best]
		    < bestrot * cube_freeness[i]
		    || cube_freeness[i] == 0.0))
	    {
		best = i;
		bestrot = rotations[i];
	    }
	}
	usedbit[best] = 1;
	bitordering[best] = bit;

#ifdef DEBUG_BITORDERING	
	printf("\n");
	int z, y, x;
	for (z = 2; z >= 0; z--) {
	    for (y = 3; y >= 0; y--) {
		for (x = 0; x < 5; x++) {
		    int i = coord2bit(x,y,z);
		    printf("%3.3lf ", cube_freeness[i]);
		}
	        printf("    ");
		for (x = 0; x < 5; x++) {
		    int i = coord2bit(x,y,z);
		    printf("%6.2lf%c ", rotations[i] / cube_freeness[i],
			   i == best ? '*': ' ');
		}
		printf("\n");
	    }
	    printf("\n");
	}
	printf(" %d", best);
#endif


	if (bestrot == 0.0)
	    continue;
	
	for (p = piece_list; p; p = p->next) {
	    for (i = 4; i < 32 && i <= best; i++) {
		long long *rots = p->rots[i];
		while (*rots != 1LL) {
		    if (*rots & (1ULL << best)) {
			double weight = 1.0;
			for (j = i; j < 64; j++) {
			    if (*rots & (1ULL << j)) {
				weight *= cube_freeness[j];
			    }
			}
			for (j = i; j < 64; j++) {
			    if (*rots & (1ULL << j)) {
				rotations_used[j] += weight;
			    }
			}
		    }
		    rots++;
		}
	    }
	    for (i = 32; i <= best; i++) {
		long *rots_hi = p->rots_hi[i-32];
		while (*rots_hi != 0) {
		    if (*rots_hi & (1UL << (best-32))) {
			double weight = 1.0;
			for (j = i; j < 64; j++) {
			    if (*rots_hi & (1UL << (j-32))) {
				weight *= cube_freeness[j];
			    }
			}
			for (j = i; j < 64; j++) {
			    if (*rots_hi & (1UL << (j-32))) {
				rotations_used[j] += weight;
			    }
			}
		    }
		    rots_hi++;
		}
	    }
	}

	double factor = cube_freeness[best] / bestrot;
	for (j = 4; j < 64; j++) {
	    cube_freeness[j] *= 1 - factor * rotations_used[j];
	}
	cube_freeness[best] = 0;
    }
#ifdef DEBUG_BITORDERING	
    printf ("\n");
#endif

    if (first_piece) {
	for (p = first_piece; p; p = p->next) {
	    free(p->rots[4]);
	}
    }

    i = 0;
    last = &first_piece;
    for (p = piece_list; p; p = p->next) {
	transpose_piece(p, &reordered[i]);
	*last = &reordered[i];
	last = &reordered[i].next;
	i++;
    }
    last = NULL;
    board = transpose(board);
#ifdef DEBUG_BITORDERING	
    printf("reordering done\n");
#endif
    return board;
}

void placeFirst(long long rot) {
    int oldsolutions = solutions;
    long long board;
    placed++;
#ifdef COUNT_PLACED
    placedP[0]++;
#endif
    /* The first piece is the x piece.  We must distinguish whether that
     * piece is in a symmetric position or not.  If it is symmetric we
     * need to use the asymmetric_piece to remove symmetric solutions.
     *
     * We also hack the used vector so that the hashtable isn't shared
     * between symmetric and asymmetric cases.
     */
    if (mirror(rot) == rot) {
	used = 1 << X_PIECE;
	compiled[ASYMMETRIC_PIECE-1].next = &asym;
    } else {
	used = 0;
	compiled[ASYMMETRIC_PIECE-1].next = &compiled[ASYMMETRIC_PIECE];
    }

    board = reorder(15ULL | rot, &compiled[0]);
    rotpieces[0] = board;

    /* hashtable mustn't be reused! */
    memset(hashtable, 0, HASH_SIZE * sizeof(*hashtable));

    solve(board, coord2bit(0,0,0), 1);
#ifdef STATISTICS
    printf("0: %2d %4d solutions ; %7d solutions total\n", 
	   X_PIECE, solutions - oldsolutions, solutions);
    dumpBoard(15ULL, board);
    statistics();
#endif
}

int main() {
    int i;
    /* compile all pieces.  Remove all symmetries for the X_PIECE */
    for (i = 0 ; i< NUM_PIECES; i++) {
	compile(i, i == X_PIECE ? 2 : 0, &compiled[i]);
    }
    /* compile ASYMMETRIC_PIECE again with removed symmetries.  It
     * is used when the X_PIECE is symmetric.
     */
    compile(ASYMMETRIC_PIECE, 1, &asym);
    
    /* link the compiled pieces together */
    for (i = 0 ; i< NUM_PIECES-1; i++)
	compiled[i].next = &compiled[i+1];
    compiled[NUM_PIECES-1].next = NULL;

    /* remove X_PIECE from the list, as it is placed first. */
    compiled[X_PIECE-1].next = &compiled[X_PIECE+1];
    /* Set asym.next to right value. */
    asym.next = &compiled[ASYMMETRIC_PIECE + 1];
    
    printf("Rotation table built\n");
    
    first_piece = NULL;
    solutions = 0;
    placed = 0;
#ifdef COUNT_PLACED
    for (i = 0; i < NUM_PIECES; i++) {
	placedP[i] = 0;
    }
#endif
    
    hashtable = malloc(HASH_SIZE * sizeof(*hashtable));
    printf("Hashtable allocated\n");
    
    /* place x piece */
    for (i = 4; i < 32; i++) {
	long long *rots = compiled[X_PIECE].rots[i];
	while (*rots != 1LL) {
	    placeFirst(*rots);
	    rots++;
	}
    }
    for (i = 32; i < 64; i++) {
	long *rots_hi = compiled[X_PIECE].rots_hi[i-32];
	while (*rots_hi != 0L) {
	    placeFirst(((long long) *rots_hi) << 32);
	    rots_hi++;
	}
    }
    printf("solutions: %d\n", solutions);
    return 0;
}
